<?php
/************************************************************************************************
 * © Copyright 2021, BizZToolz
 * Use of this code is permitted under the terms of the GNU Lesser General Public License
 * https://www.fsf.org/
 ************************************************************************************************/

if (!defined('ABSPATH')) {
    exit;
}
// Exit if accessed directly

require_once ABSPATH . '/wp-load.php';
require_once ABSPATH . '/wp-content/plugins/gravitywise/includes/common/helper.php';

class bizz_ConnectWiseClient {
    public $cw_version_file_name;
    private $api_url;
    private $cw_header;
    private $cw_client_id;
    private $cw_site;
    private $integrator_public_key;
    private $integrator_private_key;
    private $company_id;
    private $auth;

    public function __construct()
    {
        $conf = (object) get_option('gravityformsaddon_gravitywise_settings');
    
        $this->cw_client_id = $conf->cw_client_id ?? 'e2c6d3ac-2b9f-44b6-99e4-def0fbcfbe39';
        $this->cw_version_file_name =$conf->cw_version;
        $this->cw_site = $conf->cw_site;
        $this->company_id = $conf->company_id;
        $this->integrator_public_key =  $conf->integrator_public_key;
        $this->integrator_private_key =  $conf->integrator_private_key;
        $this->auth = "$this->company_id+$this->integrator_public_key:$this->integrator_private_key";
        $this->api_url = $this->initializeInstance($this->cw_site, $this->company_id);
    }

    /**
     * It will also populate the paths and components from the swagger file
     * It uses a default api file if no version is selected
     * @return string api url
     */
    public function initializeInstance($site, $companyId)
    {
        require_once ABSPATH . '/wp-content/plugins/gravitywise/includes/common/helper.php';

        // populate paths and components based on the default or selected version
        if (!empty($this->cw_version_file_name)) {
            $files = bizz_Helper::getVersionChoices();
            foreach ($files as $file) {
                if ($this->cw_version_file_name == $file['value']) {
                    $apiVersion = str_replace(bizz_CommonVariables::$apiFilenameConjunction, bizz_CommonVariables::$fileSeparator, $file['label']);
                }
            }
        } else {
            $apiVersion = bizz_CommonVariables::$cwApiDefaultVersionId;
        }

        // setup generic api header
        $this->cw_header = array(
            bizz_CommonVariables::$cwApiHeaderContentTypeTemplate,
            sprintf(bizz_CommonVariables::$cwApiHeaderClientIdTemplate, $this->cw_client_id),
            sprintf(bizz_CommonVariables::$cwApiHeaderAcceptTemplate, $apiVersion),
            sprintf(bizz_CommonVariables::$cwApiHeaderAuthTemplate, base64_encode($this->auth))
        );

        // get the specific url
        $url = sprintf(bizz_CommonVariables::$cwLoginUrl, $site, $companyId);

        if (!empty($this->auth)) {
            $data = $this->curlGetRequest($url);
        }

        if ($data == null) {
            $returnedurl = bizz_CommonVariables::$cwApiUrlError;
        } elseif ($data->Codebase == null) {
            //show display error
            $returnedurl = bizz_CommonVariables::$cwApiUrlError;
        } elseif ($data->IsCloud == true) {
            //url format for true is: https://api-na.myconnectwise.net/v2016_4/apis/3.0/
            $returnedurl = sprintf(bizz_CommonVariables::$cwCloudUrl, $site, $data->Codebase);
        } else {
            //url format for false is: https://cw.bngteam.com/v4_6_release/apis/3.0/
            $returnedurl = sprintf(bizz_CommonVariables::$cwCloudUrl, $site, $data->Codebase);
            $returnedurl = sprintf(bizz_CommonVariables::$cwNonCloudUrl, $site, $data->Codebase);
        }

        return $returnedurl;
    }

    /**
     * Retrieve connectwise data
     *
     * @param string $extension - the connectwise api endpoint
     * @param string $qsParameters - the query string parameters for the api
     *
     * @return array $data containing connectwise data
     */
    public function getCwData($extension, $qsParameters)
    {
        $serviceUrl = $this->api_url . $extension;
        if (!empty($qsParameters)) {
            $serviceUrl .= "?{$qsParameters}";
        }
        $data = $this->curlGetRequest($serviceUrl);
        return $data;
    }

    /**
     * Creates connectwise data
     *
     * @param string $extension - the connectwise api endpoint
     * @param string $args - the body of the component to be create
     *
     * @return array $data containing connectwise data or error
     */
    public function postCwData($extension, $args)
    {
        $serviceUrl = $this->api_url . $extension;
        $data = $this->curlPostRequest($serviceUrl, $args);
        return $data;
    }

    /**
     * Updates connectwise data
     *
     * @param string $extension - the connectwise api endpoint
     * @param string $args - the body of the component to be update
     *
     * @return array $data containing connectwise data
     */
    public function patchCwData($extension, $args)
    {
        $serviceUrl = $this->api_url . $extension;
        $data = $this->curlPatchRequest($serviceUrl, $args);
        return $data;
    }

    /**
     * Makes get requests using curl
     * Verifies the request, logs error and either throw an exception
     * or returns the data
     *
     * @return object data
     */
    public function curlGetRequest($serviceUrl, $args = null)
    {
        $curl = curl_init($serviceUrl);

        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_HTTPGET, 1);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $this->cw_header);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false); // added to assist localhost development

        $curl_response = curl_exec($curl);
        $curl_info = curl_getinfo($curl);
        $data = json_decode($curl_response);
        curl_close($curl);

        if ($this->validateCurlRequest($curl_info, $data, $args)) {
            return $data;
        }
    }

    /**
     * Makes delete requests using curl to connectwise API
     *
     * @return object data received
     */
    public function curlDeleteRequest($serviceUrl, $args = null)
    {
        $curl = curl_init($serviceUrl);

        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'DELETE');
        curl_setopt($curl, CURLOPT_HTTPHEADER, $this->cw_header);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false); // added to assist localhost development
        
        $curl_response = curl_exec($curl);
        $curl_info = curl_getinfo($curl);
        $data = json_decode($curl_response);
        curl_close($curl);

        if ($this->validateCurlRequest($curl_info, $data, $args)) {
            return $data;
        }
    }

    /**
     * Makes Post requests to the connectwise API
     *
     * @return object data received
     */
    public function curlPostRequest($serviceUrl, $postData, $args = null)
    {
        $curl = curl_init($serviceUrl);
        $curl_post_data = json_encode($postData);
        $headerContentLength = sprintf(bizz_CommonVariables::$cwApiHeaderContentLengthTemplate, strlen($curl_post_data));

        $headers = $this->cw_header;
        array_push($headers, $headerContentLength);

        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_POST, true);
        curl_setopt($curl, CURLOPT_POSTFIELDS, $curl_post_data);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false); // added to assist localhost development

        $curl_response = curl_exec($curl);
        $curl_info = curl_getinfo($curl);
        $data = json_decode($curl_response);
        if(curl_errno($curl))
        {
            $s = 'Curl error: ' . curl_error($curl);
            curl_close($curl);
            return ['errors'=> $s];
        }
        curl_close($curl);

        $this->validateCurlRequest($curl_info, $data, $args);
        return $data;
    }

    /**
     * Makes patch requests to the connectwise manage API
     *
     * @param object data received
     */
    public function curlPatchRequest($serviceUrl, $patchData, $args = null)
    {
        $curl = curl_init($serviceUrl);
        $curl_post_data = json_encode($patchData);
        $headerContentLength = sprintf(bizz_CommonVariables::$cwApiHeaderContentLengthTemplate, strlen($curl_post_data));

        $headers = $this->cw_header;
        array_push($headers, $headerContentLength);

        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, 'PATCH');
        curl_setopt($curl, CURLOPT_POSTFIELDS, $curl_post_data);
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        // added to assist localhost development

        $curl_response = curl_exec($curl);
        $curl_info = curl_getinfo($curl);
        $data = json_decode($curl_response);
        curl_close($curl);

        if ($this->validateCurlRequest($curl_info, $data, $args)) {
            return $data;
        }
    }

    /**
     * Validates the curl request
     */
    private function validateCurlRequest($curl_info, $data, $args = null)
    {
        if (isset($data->code)) {
            $curl_error = $curl_info['http_code'];

            if ($data->code != '') {
                //do error stuff
                $error_code = $data->code;
                $error_message = $data->message;
                $errors = "";
                if (isset($data->errors)) {
                    if (count($data->errors) > 0) {
                        for ($x = 0; $x < count($data->errors); $x++) {
                            $thiserror = $data->errors[$x];

                            if ($errors == '') {
                                $errors .= "Exact Errors Reported:";
                            } else {
                                $errors .= "\n";
                            }
                            $errors .= " - " . $thiserror->code . " (" . $thiserror->message . ")";
                        }
                    }
                }
                $error_display = "Error Logged while making a curl request: ";
                $error_display .= $error_code . ": " . $error_message . "\n" . $errors . PHP_EOL . PHP_EOL;
                $error_display .= "Url: " . $curl_info['url'] . PHP_EOL;
                if ($args != null) {
                    $error_display = $this->readDetailsFromArgs($error_display, $args);
                }
                return false;
            } elseif ($curl_error > 400) {
                //check forr curl errors
                $error_display = "Error Logged while making a curl request:: " . $curl_error . PHP_EOL . PHP_EOL;
                $error_display .= "Url: " . $curl_info['url'] . PHP_EOL;
                if ($args != null) {
                    $error_display = $this->readDetailsFromArgs($error_display, $args);
                }
                return false;
            }
        }
        else {
            //good result, return data
            return $data;
        }
    }

    /**
     * Reads site details from the args passed in the request
     */
    private function readDetailsFromArgs($error_display, $args)
    {
        $error_display .= "Site: " . $args['feed_details']['site'] . PHP_EOL;
        $error_display .= "Page: " . $args['feed_details']['page'] . PHP_EOL;
        $error_display .= "Form: " . $args['feed_details']['form'] . PHP_EOL;
        $error_display .= "Feed: " . $args['feed_details']['feed'] . PHP_EOL;
        return $error_display;
    }
}
